#include "includes.h"
#include "revinetd.h"
#include "relay_agt.h"
#include "server.h"
#include "misc.h"

static const char cvsid[] = "$Id: revinetd.c,v 1.22 2003/03/17 22:47:58 alexcv Exp $";

int
main(int argc, char **argv)
{
    /* All variables are initialized to a known invalid value so that they
       are not used un-initialized. To use them should trip tests or cause
       segfaults. */
    static struct option long_options[] = {
        /* These options set a flag. */
        {"help", no_argument, 0, 'h'},
        {"daemonize", no_argument, 0, 'd'},
        {"relay-agent", no_argument, 0, 'r'},
        {"server", no_argument, 0, 's'},
        {"server-port-one", required_argument, 0, 'l'},
        {"server-port-two", required_argument, 0, 'L'},
        {"callback-host", required_argument, 0, 'c'},
        {"target-host", required_argument, 0, 't'},
        {"verbose", no_argument, 0, 'v'},
        {"quiet", no_argument, 0, 'q'},
        {0, 0, 0, 0}};
    /* getopt_long stores the option index here. */
    int c, option_index = 0;
    pid_t pid;
   
    printf("%s - Copyright (c) 2003\n\n", *argv);
    exec_name = strdup(*argv);
    
    init_conf();

    while (1) {
        c = getopt_long(argc, argv, "hdrsl:L:c:t:qv", long_options,
                &option_index);

        /* Detect the end of the options. */
        if (c == -1)
            break;

        switch (c) {
        case 0:
            /* If this option set a flag, do nothing else now. */
            if (long_options[option_index].flag != 0)
                break;
            printf("option %s", long_options[option_index].name);
            if (optarg)
                printf(" with arg %s", optarg);
            printf("\n");
            break;

        case 'd':
            conf.daemonize = 1;
            conf.verbosity = VB_QUIET;
            break;
            
        case 's':
            if (conf.server_flag == FLAG_NONE) {
                conf.server_flag = FLAG_SERVER;
            } else {
                fprintf(stderr, "You can only set one mode, either Server or "
                                "Relay Agent, not both.\n\n");
                usage();
            }
            break;

        case 'r':
            if (conf.server_flag == FLAG_NONE) {
                conf.server_flag = FLAG_RELAY;
            } else {
                fprintf(stderr, "You can only set one mode, either Server or "
                                "Relay Agent, not both.\n\n");
                usage();
            }
            break;

        case 'l':
            if (conf.server_flag == FLAG_SERVER) {
                conf.port = atoi(optarg);
            } else {
                fprintf(stderr, "You must select Server mode to use -l or "
                                "--server-port-one.\n\n");
                usage();
            }
            break;

        case 'L':
            if (conf.server_flag == FLAG_SERVER) {
                conf.port2 = atoi(optarg);
            } else {
                fprintf(stderr, "You must select Server mode to use -L or "
                                "--server-port-two.\n\n");
                usage();
            }
            break;

        case 'c':
            if (conf.server_flag == FLAG_RELAY) {
                conf.callback = strdup(optarg);
                conf.port = parse_host_str(conf.callback);
            } else {
                fprintf(stderr, "You must select Relay Agent mode to use -c "
                                "or --callback-host.\n\n");
                usage();
            }
            break;
    
        case 't':
            if (conf.server_flag == FLAG_RELAY) {
                conf.target = strdup(optarg);
                conf.port2 = parse_host_str(conf.target);
            } else {
                fprintf(stderr, "You must select Relay Agent mode to use -t "
                                "or --target-host.\n\n");
                usage();
            }
            break;
        case 'h':
            usage();
            break;
            
        case 'v':
            if (conf.daemonize == 0) {
                conf.verbosity = VB_VERBOSE;
            }
            break;
            
        case 'q':
            conf.verbosity = VB_QUIET;
            break;
            
        case '?':
            /* FALL THROUGH */
        default:
            usage();
        }
    }

    /* Print any remaining command line arguments (not options). */
    if (optind < argc) {
        usage();
    }

    if (conf.daemonize == TRUE) {
        pid = fork();

        if (pid > 0) { /* parent */
            exit(0);
        } else if (pid < 0) { /* error */
            perror("fork");
            exit(1);
        } /* child */
        setsid();

        close(0); close(1); close(2);
    }
    
    signal(SIGTERM, clean_exit);
    signal(SIGINT, clean_exit);
    signal(SIGQUIT, clean_exit);
    signal(SIGHUP, clean_exit);
    
    if (conf.server_flag == FLAG_SERVER) {
        if (conf.port >= 0 && conf.port2 >= 0) {
            server(conf.port, conf.port2);
        } else {
            fprintf(stderr, "Two ports, port1 (-l and --server-port-one) and "
                            "port2 (-L and --server-port-two) must be "
                            "provided.\n\n");
            usage();
        }
    } else if (conf.server_flag == FLAG_RELAY) {
        if (conf.callback != NULL && conf.target != NULL) {
            relay_agent(conf.callback, conf.port, conf.target, conf.port2);
        } else {
            fprintf(stderr, "Both a callback (-c and --callback-host) and a "
                            "target (-t and --target-host)\nmust be "
                            "specified.\n\n");
            usage();
        }
    } else {
        usage();
    }

    exit(0);
}

void
usage(void)
{

    fprintf(stderr, "Usage: revinetd -s -l PORT1 -L PORT2 [-d]"
                    "\n"
                    "       revinetd -r -c HOST1:PORT1 -t "
                    "HOST2:PORT2 [-d]\n"
                    "\n"
                    "Global arguments:\n"
                    "  -s  --server\t\tRuns in server mode\n"
                    "  -r  --relay-agent\tRuns as the relay agent\n"
                    "  -d  --daemonize\tRuns as a daemon in the background\n"
                    "                 \t(implies -q)\n"
                    "  -h  --help\t\tPrints this message and exit\n"
                    "  -q  --quiet\tSurpress all messages\n"
                    "  -v  --verbose\tIncrease verbiosity\n"
                    "\n"
                    "Server arguments:\n"
                    "  -l PORT1  --server-port-one=PORT1\tClient port to listen"
                    " on\n"
                    "  -L PORT2  --server-port-two=PORT2\tRelay Agent port to"
                    " listen on\n"
                    "\n"
                    "Relay agent arguments:\n"
                    "  -c HOST1:PORT1  --callback-host=HOST1:PORT1\tHost where "
                    "the revinetd \n                                      "
                    "       \tserver instance runs\n\n"
                    "  -t HOST2:PORT2  --target-host=HOST2:PORT2\tHost where we"
                    " want to connect\n\n");

    clean_exit(SIGQUIT);
}

void
clean_exit(int sig)
{
    OpenSockets *open_sock;
    
    /* Close all registered sockets. */
    open_sock = conf.open_sock;
    while (open_sock != NULL) {
        if (shutdown(open_sock->sock, SHUT_RDWR) == -1) {
            perror("shutdown");
        }
        close(open_sock->sock);
        open_sock = open_sock->next;
    }
    
    exit(0);
}

unsigned short
parse_host_str(char *host_str)
{
    unsigned short port;
    int j = 0;
    
    /* Skip any leading spaces. */
    while (isspace(*host_str)) { host_str++; }
    
    /* Loop through the line until we reach a ':'. Exit with errors if we
       stumble on an illegal character. */
    while (*(host_str + j) != ':') {
        if (*(host_str + j) == '\0' || isspace(*(host_str + j))) {
            fprintf(stderr, "Invalid hostname format.\n");
            usage();
        }
        j++;
    }

    /* Insert a NUL byte at the position of the ':' to delimitate the
       hostname. And then skip ahead one byte. */
    *(host_str + j++) = '\0';
   
    /* Grab the port. */
    port = (unsigned short)atoi(host_str + j);

    return port;
}

void
init_conf(void)
{

    memset(&conf, 0, sizeof(Conf));
    conf.port = -1;
    conf.port2 = -1;
    conf.verbosity = VB_NORMAL;
}
